#
# Copyright (c) 2024-2025 ABXR Labs, Inc.
# Released under the MIT License. See LICENSE file for details.
#

from loguru import logger
from androguard.core.apk import APK
from androguard.util import set_log

set_log("WARNING")

def get_package_name(apk_path):
    try:
        apk = APK(apk_path)
        return apk.get_package()
    except Exception as e:
        return None




#
# Copyright (c) 2024-2025 ABXR Labs, Inc.
# Released under the MIT License. See LICENSE file for details.
#

import re
import subprocess
import time
import json
from pathlib import Path
import tempfile
import zipfile

from abxr.apk import get_package_name

from abxr.configuration_package import AbxrClientApk, ConfigurationPackage

# create custom exception for apk install failure
class DeviceInstallException(Exception):
    pass

class DeviceUninstallException(Exception):
    pass

class DeviceConfigureException(Exception):
    pass

class DeviceSetDeviceOwnerException(Exception):
    pass

class DeviceSetEnvironmentException(Exception):
    pass

class DeviceRemoveDeviceOwnerException(Exception):
    pass

class DeviceConfigureApiTokenException(Exception):
    pass

class DeviceConfigurationObjectException(Exception):
    pass



class Version:
    def __init__(self, version_str):
        parts = version_str.split('.')
        self.major = int(parts[0]) if len(parts) > 0 and parts[0].isdigit() else 0
        self.minor = int(parts[1]) if len(parts) > 1 and parts[1].isdigit() else 0
        self.build = int(parts[2]) if len(parts) > 2 and parts[2].isdigit() else 0
        self.revision = int(parts[3]) if len(parts) > 3 and parts[3].isdigit() else 0

    def __repr__(self):
        return f"Version(major={self.major}, minor={self.minor}, build={self.build}, revision={self.revision})"


class SystemVersion:
    @staticmethod
    def sanitize(version):
        sanitized = re.sub(r'[^0-9.]', '', version)
        parts = sanitized.split('.')
        parts = parts[:4]
        
        return '.'.join(parts)

    @staticmethod
    def parse(version):
        try:
            sanitized_version = SystemVersion.sanitize(version)
            version_obj = Version(sanitized_version)
            return version_obj
        except Exception:
            return None

class BootsrapViaAuthenticationToken:
    def __init__(self, token, group, configurations):
        self.secret_value = None
        self.device_id = None
        self.enrollment_method = None
        self.configurations = configurations
        self.token = token
        self.group = group

    def __repr__(self):
        return f"BootstrapViaAuthenticationToken(token={self.token})"
    
   # adb shell dumpsys activity service xrdm.adb.AdbService 
            # '\{
            #   \"type\":\"BootstrapViaAuthenticationToken\",
            #   \"input\":\{
            #       \"token\":\{
            #           \"secretValue\":\"46261\|HQttgqpkbZa02R4XaFP7noe9SGv51n1EGH72AO5cda90c9cb\"
            #       \},
            #       \"deviceId\":\"e1eeea37-1356-4718-93aa-ac5eda69dda2\",
            #       \"enrollmentMethod\":\"admin\",
            #       \"configurations\":\{
            #           \"WifiNetworks\":\{
            #               \"networks\":[]
            #           \}
            #       \}
            #   \}
            # \}'

    def to_dict(self):
        return {
            "type": "BootstrapViaAuthenticationToken",
            "input": {
                "token": {
                    "secretValue": self.secret_value
                },
                "deviceId": self.device_id,
                "enrollmentMethod": "admin",
                "configurations": self.configurations 
            }
        }
    
    

class Device:
    def __init__(self, serial):
        self.serial = serial

        self.manufacturer_name = None
        self.model_name = None
        self.firmware_version = None
        self.system_version = None
        self.abi = None

        self._load_manufacturer_name()
        self._load_model_name()
        self._load_firmware_and_system_version()
        self._load_abi()

    def is_quest(self):
        return "Quest" in self.model_name
    
    def is_vive_flow(self):
        return "VIVE Flow" in self.model_name
    
    def is_pico(self):
        return "Pico" in self.model_name
    
    def is_vive_xr_elite(self):
        return "VIVE XR Series" in self.model_name
    
    def is_vive_focus_3(self):
        return "Focus 3" in self.model_name
    
    def is_magic_leap_2(self):
        return "Magic Leap 2" in self.model_name
    
    def prop(self, prop_name):
        result = subprocess.run(["adb", "-s", self.serial, "shell", "getprop", prop_name], capture_output=True, text=True)
        return result.stdout.strip()

    def _load_manufacturer_name(self):
        self.manufacturer_name = self.prop("ro.product.manufacturer")

    def _load_model_name(self):
        model_name = self.prop("pxr.vendorhw.product.model")
        
        if not model_name:
            model_name = self.prop("ro.product.model")
        
        self.model_name = model_name
    
    def _load_firmware_and_system_version(self):
        if self.is_vive_flow() or self.is_vive_xr_elite() or self.is_vive_focus_3():
            self.firmware_version = self.prop("ro.product.version")
        elif self.is_magic_leap_2():
            self.firmware_version = self.prop("ro.build.version.lumin")
        elif self.is_quest():
            self.firmware_version = self.get_installed_version("com.oculus.vrshell")
        else:
            self.firmware_version = self.get_installed_version("com.htc.mobilevr.launcher")

        if not self.firmware_version:
            self.firmware_version = self.prop("ro.build.display.id")

        self.system_version = SystemVersion.parse(self.firmware_version)

    def _load_abi(self):
        self.abi = self.prop("ro.product.cpu.abi")
    
    def is_abxr_provisioned(self):
        if self.abxr_client_version() is not None:
            return True
        return False
    
    def mkdir(self, path):
        proc = subprocess.run(["adb", "-s", self.serial, "shell", "mkdir", "-p", path])
        if proc.returncode:
            raise Exception(f"Failed to create directory {path} on device {self.serial}")
    
    def install_apk_obb_package(self, package_name):
        if type(package_name) == str:
            package_name = Path(package_name)

        if not package_name.suffix == ".zip":
            print(f"Skipping non-ZIP packaged file. {package_name}")
            raise DeviceInstallException(f"non-ZIP packaged file. {package_name}")
        
        with tempfile.TemporaryDirectory() as extract_path:
            try:
                zipfile.ZipFile(package_name, 'r').extractall(extract_path)
                package_name = None
                for file in Path(extract_path).rglob("*.apk"):
                    if not file.name.startswith("."):
                        package_name = get_package_name(file)
                        self.install_apk(file)

                for file in Path(extract_path).rglob("*.obb"):
                    if not file.name.startswith("."):
                        print(f"Installing OBB: {file}")
                        self.mkdir(f"/sdcard/Android/obb/{package_name}")
                        self.push_file(file, f"/sdcard/Android/obb/{package_name}/{file.name}")        
            
            except NotImplementedError:
                print(f"Skipping invalid zip package format. {package_name}")            
        
    
    def install_apk(self, package_name):
        if type(package_name) == str:
            package_name = Path(package_name)

        if not package_name.suffix == ".apk":
            print(f"Skipping non-APK file. {package_name}")
            raise DeviceInstallException(f"non-APK file. {package_name}")

        print(f"Installing APK: {package_name}")
        proc = subprocess.run(["adb", "-s", self.serial, "install", "-r", "-t", "-g", package_name], capture_output=True, text=True)
        if proc.returncode:
            raise DeviceInstallException(f"Failed to install {package_name} on device {self.serial} - {proc.stderr}")
        
    def push_file(self, local_file, remote_file):
        proc = subprocess.run(["adb", "-s", self.serial, "push", local_file, remote_file])
        if proc.returncode:
            raise DeviceInstallException(f"Failed to push {local_file} to {remote_file} on device {self.serial}")
        
    def set_google_play_protect(self, enable):
        if enable:
            proc = subprocess.run(["adb", "-s", self.serial, "shell", "settings", "put", "global", "verifier_verify_adb_installs", "1", "&&", "settings", "put", "global", "package_verifier_enable", "1"])
        else:
            proc = subprocess.run(["adb", "-s", self.serial, "shell", "settings", "put", "global", "verifier_verify_adb_installs", "0", "&&", "settings", "put", "global", "package_verifier_enable", "0"])
        
        if proc.returncode:
            raise Exception("Failed to toggle Google Play Protect.")
        
    def set_quest_sleep_enabled(self, enable):
        if enable:
            proc = subprocess.run(["adb", "-s", self.serial, "shell", "am", "broadcast", "-a", "com.oculus.vrpowermanager.automation_disable", "&&", "svc", "power", "stayon", "false"])
        else:
            proc = subprocess.run(["adb", "-s", self.serial, "shell", "svc", "power", "stayon", "true", "&&", "am", "broadcast", "-a", "com.oculus.vrpowermanager.prox_close"])   
        
        if proc.returncode:
            raise Exception("Failed to toggle Quest sleep.")
        
    def set_quest_guardian_enabled(self, enable):
        if enable:
            proc = subprocess.run(["adb", "-s", self.serial, "shell", "setprop", "debug.oculus.guardian_pause", "0"])
        else:
            proc = subprocess.run(["adb", "-s", self.serial, "shell", "setprop", "debug.oculus.guardian_pause", "1"])
        
        if proc.returncode:
            raise Exception("Failed to toggle Guardian.")
    
    def device_sleep(self, seconds):
        proc = subprocess.run(["adb", "-s", self.serial, "shell", "sleep", str(seconds)])
        if proc.returncode:
            raise Exception("Failed to sleep device.")
        
    def send_keycode_and_sleep(self, keycode, seconds):
        proc = subprocess.run(["adb", "-s", self.serial, "shell", "input", "keyevent", str(keycode)])
        if proc.returncode:
            raise Exception("Failed to send keycode.")
        
        self.device_sleep(seconds)

    def set_package_enabled(self, package_name, enable):
        if enable:
            proc = subprocess.run(["adb", "-s", self.serial, "shell", "pm", "disable-user", "--user 0", package_name])
        else:
            proc = subprocess.run(["adb", "-s", self.serial, "shell", "pm", "enable", "--user 0", package_name])
        
        if proc.returncode:
            raise Exception(f"Failed to toggle {package_name} enablement.")

    def delete_quest_accounts(self):
        home = (3, 3)
        up = (19, 0.5)
        down = (20, 0.5)
        right = (22, 0.5)
        enter = (66, 0.5)

    
        requires_extra_down = False
        if self.system_version and self.system_version.major >= 62:
            requires_extra_down = True

        account_delete_steps = [
            home,
            "am start -n 'com.android.settings/.Settings$AccountDashboardActivity'",
            "sleep 5",
            down,
            up,
            down,
        ]

        if requires_extra_down:
            account_delete_steps.append(down)

        account_delete_steps += [
            enter,
            down,
            down,
            enter,
            down,
            right,
            enter,
            enter,
            down,
            down,
            enter,
            down,
            right,
            enter,
            enter,
            down,
            down,
            enter,
            down,
            right,
            enter,
            enter,
            down,
            down,
            enter,
            down,
            right,
            enter,
            "sleep 5",
        ]

        for step in account_delete_steps:
            if type(step) == tuple:
                self.send_keycode_and_sleep(*step)
            elif type(step) == str:
                proc = subprocess.run(["adb", "-s", self.serial, "shell", step])
                if proc.returncode:
                    raise Exception(f"Failed to execute step: {step}")
            else:
                raise Exception(f"Invalid step: {step}")


    def send_configuration_package(self, config_pkg):
        if not type(config_pkg) == ConfigurationPackage:
            raise Exception("You must provide a ConfigurationPackage object.")
        
        if type(config_pkg) == str|Path:
            config_pkg = ConfigurationPackage(config_pkg)

        if config_pkg.is_xrdm2:
            client_settings_file_location = "/sdcard/Setup/settings.json"
        else:
            client_settings_file_location = "/sdcard/ArborXR/Settings/settings.json"

        for app in config_pkg.apps:
            with tempfile.TemporaryDirectory() as extract_path:
                item = config_pkg.extract_item(app, extract_path)
                self.install_apk(item)

        for obb_packaged_app in config_pkg.obb_packaged_apps:
            with tempfile.TemporaryDirectory() as extract_path:
                item = config_pkg.extract_item(obb_packaged_app, extract_path)
                self.install_apk_obb_package(item)

        for file in config_pkg.files:
            with tempfile.TemporaryDirectory() as extract_path:
                item = config_pkg.extract_item(file, extract_path)
                target_file_location = file.split("/files")[-1]
                self.push_file(item, target_file_location)

        for video in config_pkg.videos:
            with tempfile.TemporaryDirectory() as extract_path:
                item = config_pkg.extract_item(video, extract_path)
                target_file_location = video.split("/videos")[-1]
                self.push_file(item, f"/sdcard/ArborXR/videos{target_file_location}")

        for launcher_media in config_pkg.launcher_media:
            with tempfile.TemporaryDirectory() as extract_path:
                item = config_pkg.extract_item(launcher_media, extract_path)
                target_file_location = launcher_media.split("/launcher-media")[-1]
                self.push_file(item, f"/sdcard/launcher-media{target_file_location}")

        for images in config_pkg.images:
            with tempfile.TemporaryDirectory() as extract_path:
                item - config_pkg.extract_item(images, extract_path)
                target_file_location = images.split("/images")[-1]
                self.push_file(item, f"/sdcard/ArborXR/images{target_file_location}")
        

        with tempfile.NamedTemporaryFile() as f:
            f.write(config_pkg.read_settings_json_from_zip())
            f.flush()
            self.push_file(f.name, client_settings_file_location)
        
        if config_pkg.is_xrdm2:
            command = {
                "type": "BootstrapViaJsonFile",
                "input": {
                    "file": client_settings_file_location
                }
            }
            proc = subprocess.run(["adb", "-s", self.serial, "shell", "dumpsys", "activity", "service", "xrdm.adb.AdbService",  f"'''{command}'''"])
            if proc.returncode:
                raise DeviceConfigureException("Failed to configure device with settings.json")
        else:
            proc = subprocess.run(["adb", "-s", self.serial, "shell", "am", "broadcast", "-n", "app.xrdm.client/.SetupDeviceReceiver", "-a", "CONFIGURE_DEVICE", "-e", "ClientSettingsFile", client_settings_file_location])
            if proc.returncode:
                raise DeviceConfigureException("Failed to configure device with settings.json")

    def force_stop_and_sleep(self, package_name, seconds):
        proc = subprocess.run(["adb", "-s", self.serial, "shell", "am", "force-stop", package_name])
        if proc.returncode:
            raise Exception(f"Failed to force stop {package_name} on device {self.serial}")
        
        self.device_sleep(seconds)

    def reboot(self):
        proc = subprocess.run(["adb", "-s", self.serial, "reboot"])
        if proc.returncode:
            raise Exception(f"Failed to reboot device {self.serial}")

    def set_device_owner(self):
        proc = subprocess.run(["adb", "-s", self.serial, "shell", "dpm", "set-device-owner", "app.xrdm.client/.AdminReceiver"], capture_output=True, text=True)
        if proc.returncode:
            if "Not allowed to set the device owner because there are already some accounts on the device" in proc.stderr:
                if self.is_quest():
                    self.force_stop_and_sleep("com.oculus.shellenv", 10.0)
                    self.delete_quest_accounts()
                    self.set_device_owner()

                elif self.is_vive_flow():
                    htcLauncherPackageName = "com.htc.vrs.launcher"

                    self.set_package_enabled(htcLauncherPackageName, False)
                    time.sleep(1)
                    self.set_device_owner()
                    self.set_package_enabled(htcLauncherPackageName, True)
                    self.send_keycode_and_sleep(3, 0.0)

            else:
                raise DeviceSetDeviceOwnerException("Failed to set device owner." + proc.stderr)

    def set_production_environment(self):
        proc = subprocess.run(["adb", "-s", self.serial, "shell", "echo production > /sdcard/.xrdm-environment"])
        if proc.returncode:
            raise DeviceSetEnvironmentException("Failed to set production environment.")

    def remove_device_owner(self):
        proc = subprocess.run(["adb", "-s", self.serial, "shell", "am", "broadcast", "-n", "app.xrdm.client/.AdminRemover", "-a", "REMOVE_DEVICE_OWNER"])
        max_wait_sec = 10
        wait_sec = 0
        if proc.returncode == 0:
            while True and wait_sec < max_wait_sec:
                proc = subprocess.run(["adb", "-s", self.serial, "shell", "dumpsys", "device_policy", "|", "grep", "\"app.xrdm.client/.AdminReceiver\""], capture_output=True, text=True)
                if proc.stdout == "":
                    break
                print("Waiting for device owner removal...")
                time.sleep(1)
                wait_sec += 1
            
            print("Device owner removal complete.")
            return True
        raise DeviceRemoveDeviceOwnerException("Failed to remove device owner.")

    def uninstall_apk(self, package_name):
        proc = subprocess.run(["adb", "-s", self.serial, "uninstall", package_name], capture_output=True, text=True)
        if proc.returncode:
            raise DeviceUninstallException(f"Failed to uninstall {package_name} on device {self.serial} - {proc.stderr}")

    def configure_api_token_authentication(self, config_pkg):
        if not type(config_pkg) == ConfigurationPackage:
            raise Exception("You must provide a ConfigurationPackage object.")
        
        if type(config_pkg) == str|Path:
            config_pkg = ConfigurationPackage(config_pkg)

        settings = config_pkg.settings

        if settings.is_xrdm2:
            token = settings["token"]
            group = settings["group"]
            configurations = settings["configurations"]

            token = BootsrapViaAuthenticationToken(token, group, configurations)

            token_json = json.dumps(token.to_dict())

            proc = subprocess.run(["adb", "-s", self.serial, "shell", "dumpsys", "activity", "service", "xrdm.adb.AdbService",  f"'''{token_json}'''"])

            if proc.returncode:
                raise DeviceConfigurationObjectException("Failed to configuration device with object from config package")
        else:
            api_token_auth = {}
            api_token_auth["apiToken"] = settings["apiToken"]
            api_token_auth["deviceGroup"] = {}
            api_token_auth["deviceGroup"]["id"] = settings["deviceGroup"]["id"]
            api_token_auth_json = json.dumps(api_token_auth)

            proc = subprocess.run(["adb", "-s", self.serial, "shell", "am", "broadcast", "-n", "app.xrdm.client/.SetupDeviceReceiver", "-a", "CONFIGURE_API_TOKEN", "-e", "ApiTokenAuth", f"'''{api_token_auth_json}'''"])
            if proc.returncode:
                raise DeviceConfigureApiTokenException("Failed to configure device with API token authentication.")
            
    def get_installed_version(self, package_name):
        result = subprocess.run(["adb", "-s", self.serial, "shell", "dumpsys", "package", package_name], capture_output=True, text=True)
        if result.returncode != 0:
            return None
        
        for line in result.stdout.split("\n"):
            if "versionName=" in line:
                return line.split("=")[1].strip()
    
        return None
    
    def abxr_client_version(self):
        return self.get_installed_version(AbxrClientApk.PACKAGE_NAME)
    
    def is_packaged_installed(self, package_name):
        return self.get_installed_version(package_name) is not None
    
    def __str__(self):
        return self.serial




#
# Copyright (c) 2024-2025 ABXR Labs, Inc.
# Released under the MIT License. See LICENSE file for details.
#

import json
from pathlib import Path
import zipfile

class AbxrClientApk:
    PACKAGE_NAME = "app.xrdm.client"

    def __init__(self, path):
        if type(path) == str:
            path = Path(path)

        if not path.exists():
            raise Exception(f"Error: {path} does not exist.")
        
        self.path = path
        self.version = None

        self._load_version()

    def _load_version(self):
        self.version = self.get_client_version_from_filename(self.path)

    def get_client_version_from_filename(self, apk_filename):
        return apk_filename.name.split('-')[-1].split('.apk')[0]
    
class AbxrLauncherApk:
    PACKAGE_NAME = "app.xrdm.launcher"

    def __init__(self, path):
        if type(path) == str:
            path = Path(path)

        if not path.exists():
            raise Exception(f"Error: {path} does not exist.")
        
        self.path = path
        self.version = None

        self._load_version()

    def _load_version(self):
        self.version = self.get_launcher_version_from_filename(self.path)

    def get_launcher_version_from_filename(self, apk_filename):
        return apk_filename.name.split('-')[-1].split('.apk')[0]


class ConfigurationPackage:
    def __init__(self, path):
        if type(path) == str:
            path = Path(path)

        self.path = path
        self.version = None
        self.is_xrdm2 = False
        self.settings = None
        self.zip_file = None
        self.client_apk = None
        self.launcher_apk = None

        self.apps = []
        self.obb_packaged_apps = []
        self.files = []
        self.videos = []
        self.launcher_media = []

        self._open()
        self._load_settings_json()

        self._find_client_apk()
        self._find_launcher_apk()
        self._find_apps()
        self._find_files()
        self._find_videos()
        self._find_launcher_media()
        self._find_images()

        self._split_obb_packaged_apps_from_apps()

    def _open(self):
        try:
            self.zip_file = zipfile.ZipFile(self.path, 'r')
        except zipfile.BadZipFile:
            raise Exception(f"Error: {self.path} is not a valid zip file")

    def _load_settings_json(self):
        try:
            self.settings = json.loads(self.read_settings_json_from_zip().decode('utf-8'))
            self.version = self.settings.get("version", None)
            if self.version == "2.0.0":
                self.is_xrdm2 = True
        except Exception as e:
            raise Exception(f"Error parsing settings.json: {e}")
        
    def _find_client_apk(self):
        client_apk = [f for f in self.zip_file.namelist() if 'client/app.xrdm.client' in f]
        
        if len(client_apk) > 1:
            raise Exception("Multiple ArborXR APK files found in the zip file.")
        elif len(client_apk) == 0:
            raise Exception("ArborXR Client APK not found in the zip file.")

        self.client_apk = client_apk[0]

    def _find_launcher_apk(self):
        launcher_apk = [f for f in self.zip_file.namelist() if 'launcher/app.xrdm.launcher' in f]

        if len(launcher_apk) > 1:
            raise Exception("Multiple ArborXR APK files found in the zip file.")
        elif len(launcher_apk) == 1:         
            self.launcher_apk = launcher_apk[0]
        else:
            self.launcher_apk = None

    def _find_apps(self):
        self.apps = [f for f in self.zip_file.namelist() if 'apps/' in f]

    def _find_files(self):
        self.files = [f for f in self.zip_file.namelist() if 'files/' in f]

    def _find_videos(self):
        self.videos = [f for f in self.zip_file.namelist() if 'videos/' in f]

    def _find_launcher_media(self):
        self.launcher_media = [f for f in self.zip_file.namelist() if 'launcher-media/' in f]

    def _find_images(self):
        self.images = [f for f in self.zip_file.namelist() if 'images/' in f]

    def _split_obb_packaged_apps_from_apps(self):
        self.obb_packaged_apps = [f for f in self.apps if '.zip' in f]
        self.apps = [f for f in self.apps if '.zip' not in f]

    def created_at(self):
        try:
            return self.settings["createdAt"]
        except Exception as e:
            print("Error getting created at: {e}")
            return None

    def device_group(self):
        try:
            if self.is_xrdm2:
                return self.settings["group"]
            else:
                return self.settings["deviceGroup"]
        except Exception as e:
            print("Error getting device group: {e}")
            return None

    def model_name(self):
        try:
            return self.settings["deviceModel"]["name"]
        except Exception as e:
            print("Error getting device model name: {e}")
            return None
        
    def model_names(self):
        try:
            if self.is_xrdm2:
                return self.settings["deviceModel"]["enrollmentIdentifiers"]
            else:
                return self.settings["deviceModel"]["modelNames"]
        except Exception as e:
            print("Error getting device model names: {e}")
            return None
        
    def extract_item(self, item, output_path):
        try:
            path = self.zip_file.extract(item, output_path)
            return path
        except Exception as e:
            raise Exception(f"Error extracting {item} to {output_path}: {e}")
        
    def extract_client_apk_from_zip(self, output_path):        
        try:
            path = self.extract_item(self.client_apk, output_path)
            return AbxrClientApk(path)
        except Exception as e:
            raise Exception(f"Error extracting ArborXR APK from zip: {e}")
    
    def read_settings_json_from_zip(self):
        try:
            settings_json_locations = [f for f in self.zip_file.namelist() if 'settings.json' in f]
            if len(settings_json_locations) > 1:
                raise Exception("Multiple settings.json files found in the zip file.")
            
            settings_json_location = settings_json_locations[0]

            if settings_json_location:
                with self.zip_file.open(settings_json_location) as settings_file:
                    return settings_file.read()
            else:
                raise Exception("settings.json not found in the zip file.")
        except Exception as e:
            raise Exception(f"Error reading settings.json from zip: {e}")
        


#
# Copyright (c) 2024-2025 ABXR Labs, Inc.
# Released under the MIT License. See LICENSE file for details.
#

import io
import time
from abxr.version import version

i2c = None
disp = None

def is_raspberrypi():
    try:
        with io.open('/sys/firmware/devicetree/base/model', 'r') as m:
            if 'raspberry pi' in m.read().lower(): return True
    except Exception:
        pass
    
    return False


if is_raspberrypi():
    from board import SCL, SDA
    import busio
    import adafruit_ssd1306

    i2c = busio.I2C(SCL, SDA)
    disp = adafruit_ssd1306.SSD1306_I2C(128, 32, i2c)


class Display:
    LINE_HEIGHT = 8
    MAX_CHARACTERS_PER_LINE = 21

    LINE_POSITIONS = [
        LINE_HEIGHT * 0,
        LINE_HEIGHT * 1,
        LINE_HEIGHT * 2,
        LINE_HEIGHT * 3
    ]

    def __init__(self):
        self.disp = disp
        self.font_file = "abxr/font5x8.bin"
        self.line_height = 8

    def show(self):
        if self.disp:
            self.disp.show()

    def clear(self):
        if self.disp:
            self.fill(0)
            self.show()

    def text(self, text, x, y, color):
        if self.disp:
            self.disp.text(text, x, y, color, font_name=self.font_file)

    def fill(self, color):
        if self.disp:
            self.disp.fill(color)

    def show(self):
        if self.disp:
            self.disp.show()

    def write(self, text, wait=0):
        if self.disp is None:
            return

        self.clear()

        if type(text) is str:
            text = [ text ]

        wrapped_text = []
        for line in text:
            if len(line) > self.MAX_CHARACTERS_PER_LINE:
                for i in range(0, len(line), self.MAX_CHARACTERS_PER_LINE):
                    wrapped_text.append(line[i:i + self.MAX_CHARACTERS_PER_LINE].strip())
            else:
                wrapped_text.append(line)

        current_line = 0
        for line in wrapped_text:
            if current_line >= len(self.LINE_POSITIONS):
                break
            self.text(line, 0, self.LINE_POSITIONS[current_line], 255)
            current_line += 1
            
        self.show()

        if wait:
            time.sleep(wait)

   


display = Display()




#!/usr/bin/env python3
#
# Copyright (c) 2024-2025 ABXR Labs, Inc.
# Released under the MIT License. See LICENSE file for details.
#

import argparse
import os
import platform
import subprocess
import tempfile
import time

from pathlib import Path
from threading import Semaphore

from abxr.configuration_package import AbxrClientApk, AbxrLauncherApk, ConfigurationPackage
from abxr.device import Device, \
                        DeviceInstallException, \
                        DeviceUninstallException, \
                        DeviceConfigureException, \
                        DeviceSetDeviceOwnerException, \
                        DeviceSetEnvironmentException, \
                        DeviceRemoveDeviceOwnerException, \
                        DeviceConfigureApiTokenException
from abxr.display import display
from abxr.version import version


ASSETS_PATH = os.environ.get("ABXR_ASSETS_PATH", Path.home() / "abxr")
CONFIG_PACKAGE_PATH = ASSETS_PATH / Path("configuration-packages")
CLIENT_OVERRIDE_PATH = ASSETS_PATH / Path("client")

usb_event_handled = 0
usb_semaphore = Semaphore(0)
hotplug = False

def get_client_apk_override(args):
    if args.client:
        return AbxrClientApk(Path(args.client))

    for f in CLIENT_OVERRIDE_PATH.glob("*.apk"):
        return AbxrClientApk(f)
    
    return None


def get_connected_devices(args):
    result = subprocess.run(["adb", "devices"], capture_output=True, text=True)
    lines = result.stdout.strip().split("\n")[1:]
    serial_numbers = [line.split("\t")[0] for line in lines if "device" in line]

    devices = []

    for serial in serial_numbers:
        devices.append(Device(serial))

    if args.device_serial:
        devices = [d for d in devices if d.serial == args.device_serial]

    return devices


def get_configuration_packages(args):
    packages = []
    models = []

    if args.config_package:
        try:
            packages.append(ConfigurationPackage(args.config_package))
        except Exception as e:
            print(f"Error loading configuration package: {e}")
    else:
        for f in CONFIG_PACKAGE_PATH.glob("*.zip"):
            try:
                package = ConfigurationPackage(f)
                if package.model_name() in models:
                    print(f"Duplicate configuration package detected for {package.model_name()} at {package.path}")
                    print("Automatic device package selection is unavailable.")
                    return []

                print(f"Configuration package for {package.model_name()}, {package.path.name} ({package.device_group()['title']}), created: {package.created_at()}")
                packages.append(package)
                models.append(package.model_name())
            except Exception as e:
                print(f"Error loading configuration package: {e}")
                continue

    return packages


def get_configuration_package_for_device(device, packages):
    print(f"Finding configuration package for device: {device.model_name}")

    for package in packages:
        if package.model_name() == device.model_name:
            return package
        elif package.model_name() == device.model_name.replace("-", " "):
            return package
        elif device.model_name in package.model_names():
            return package
        
    return None


def trigger_device_plug_event():
    global usb_event_handled
    if usb_event_handled == 0:
        usb_event_handled = 1
        print("Waking up due to USB plug event...")
        usb_semaphore.release()


def usb_event_observer(action, device):
    if action == "add":
        trigger_device_plug_event()


def provision(args):
    print("ArborXR Field Provisioning Tools")
    print(f"Version: {version}")
 
    display.write([ "ArborXR", "Field Provisioning", f"Version: {version}"], wait=5)

    devices = get_connected_devices(args)
    if len(devices) > 0:
        usb_semaphore.release()

    packages = get_configuration_packages(args)
    if len(packages) == 0:
        print("No configuration packages found.")
        display.write("No configuration packages found.")
        if not hotplug:
            exit(0)

    number_of_config_packages = len(packages)

    while True:
        print("Waiting for device to be connected...")
        display.write([f"{number_of_config_packages} packages loaded", "", "Waiting for device to be connected."])

        global usb_event_handled
        usb_event_handled = 0
        usb_semaphore.acquire()
 
        if hotplug:
            print("Waiting for ADB device detection to settle...")
            display.write(["Waiting for ADB", "device detection."])
            time.sleep(5)
        
        devices = get_connected_devices(args)
        if len(devices) == 0:
            print("No devices connected.")
            display.write("No devices connected.", wait=5)
            if not hotplug:
                exit(0)
            continue

        for device in devices:
            package = get_configuration_package_for_device(device, packages)
            if package is None:
                print(f"No matching configuration package found for device model: {device.model_name} sn: {device.serial}")
                display.write([f"No package found for device: {device.serial}"], wait=5)
                continue
            
            print(f"Using configuration package: {package.path}")
            
            timer = time.time()

            try:
                with tempfile.TemporaryDirectory() as tempdir:
                    if not args.package_only:
                        client_apk = get_client_apk_override(args) or package.extract_client_apk_from_zip(tempdir)
                        
                        print("Using ArborXR client APK:", client_apk.path)
                        display.write(["Provisioning device:", device.model_name, device.serial, f"XRDM: {client_apk.version}"])

                        if device.is_quest():
                            device.set_quest_sleep_enabled(False)
                            device.set_quest_guardian_enabled(False)

                        device.set_google_play_protect(False)

                        #
                        # if we need to replace the client OR fresh provision    
                        #
                        if device.abxr_client_version() != client_apk.version or args.force_upgrade:
                            if device.is_packaged_installed(AbxrClientApk.PACKAGE_NAME):
                                device.remove_device_owner()
                                device.uninstall_apk(AbxrClientApk.PACKAGE_NAME)
                            
                            device.install_apk(client_apk.path)

                            if device.abxr_client_version() != client_apk.version:
                                print(f"Failed to install ArborXR Client APK on {device}")
                                display.write("Failed to install ArborXR Client APK", wait=5)
                                continue

                            device.set_device_owner()
                            device.set_production_environment()    
                                
                        else:
                            print(f"Device {device} already has the expected ArborXR Client APK version {client_apk.version}")

                    if package.launcher_apk:
                        if device.is_packaged_installed(AbxrLauncherApk.PACKAGE_NAME):
                            device.uninstall_apk(AbxrLauncherApk.PACKAGE_NAME)
                        device.install_apk(package.extract_item(package.launcher_apk, tempdir))

                    device.send_configuration_package(package)

                    if not args.package_only:
                        device.configure_api_token_authentication(package)

                display.write(["Device provisioned!", device.serial], wait=2)
                        
            except DeviceRemoveDeviceOwnerException as e:
                print(f"Failed to remove device owner: {e}")
                display.write("Failed to remove device owner", wait=5)

            except DeviceUninstallException as e:
                print(f"Failed to uninstall apk: {e}")
                display.write("Failed to uninstall apk", wait=5)

            except DeviceInstallException as e:
                print(f"Failed to install apk: {e}")
                display.write("Failed to install apk", wait=5)

            except DeviceSetDeviceOwnerException as e:
                print(f"Failed to set device owner: {e}")
                display.write("Failed to set device owner", wait=5)

            except DeviceSetEnvironmentException as e:
                print(f"Failed to set environment: {e}")
                display.write("Failed to set environment", wait=5)

            except DeviceConfigureException as e:
                print(f"Failed to configure device: {e}")
                display.write("Failed to configure device", wait=5)

            except DeviceConfigureApiTokenException as e:
                print(f"Failed to configure API token: {e}")
                display.write("Failed to configure API token", wait=5)

            except Exception as e:
                print(f"Unhandled Error: {e}")
                display.write("Unhandled Error", wait=5)

            finally:
                if device.is_quest():
                    device.set_quest_sleep_enabled(True)
                    device.set_quest_guardian_enabled(True)
                
                device.set_google_play_protect(True)

                elapsed_time_sec = time.time() - timer
                elapsed_time_min = elapsed_time_sec // 60
                elapsed_time_sec = elapsed_time_sec % 60

                print(f"Provisioning complete for {device} in {elapsed_time_min:.0f}m {elapsed_time_sec:.0f}s")

        else:
            print("No more devices to provision.")
            print("\n")
            display.write(["No more devices", "to provision."], wait=2)
            
        if not hotplug:
            exit(0)


def main():
    parser = argparse.ArgumentParser(description="ArborXR Field Provisioning Tools")
    parser.add_argument("-p", "--config-package", help="Path to configuration packages", type=str)
    parser.add_argument("-o", "--package-only", help="Only install configuration package do not provision", action="store_true")
    parser.add_argument("-c", "--client", help="Path to client APK for override", type=str)
    parser.add_argument("-s", "--device-serial", help="Serial number of device to provision", type=str)
    parser.add_argument("-f", "--force-upgrade", help="Force client install even if already installed", action="store_true")
    parser.add_argument("-l", "--list-connected-devices", help="List connected devices", action="store_true")
    args = parser.parse_args()

    if not Path(ASSETS_PATH).exists():
        parser.print_usage()
        print(f"Error: ABXR_ASSETS_PATH is not set to a valid path: {ASSETS_PATH}")
        exit(1)

    if platform.system() == "Linux":        
        import pyudev
        global hotplug

        context = pyudev.Context()
        monitor = pyudev.Monitor.from_netlink(context)
        monitor.filter_by(subsystem='usb')

        observer = pyudev.MonitorObserver(monitor, usb_event_observer)
        observer.start()

        hotplug = True
    else:
        usb_semaphore.release()

    if args.list_connected_devices:
        devices = get_connected_devices(args)
        for device in devices:
            print(device)
        exit(0)

    provision(args)

if __name__ == "__main__":
    main()



#
# Copyright (c) 2024 ABXR Labs, Inc.
# Proprietary and confidential. All rights reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
#

from setuptools import setup

setup(
    name = 'abxrfpt',
    version = '0.5.0',
    packages = ['abxr'],
    entry_points = {
        'console_scripts': [
            'abxr-provision = abxr.provision:main',
            'abxr-display-setup-mode = abxr.setup_mode:main',
        ]
    },
    install_requires = [
        'androguard',
        'pyudev',
        'adafruit-circuitpython-ssd1306',
        'setuptools'
    ],
    include_package_data=True,
)



#!/usr/bin/env python3
#
# Copyright (c) 2024 ABXR Labs, Inc.
# Proprietary and confidential. All rights reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
#


from abxr.display import display
from abxr.version import version

def main():
    display.write([ "ArborXR", "Field Provisioning", f"Version: {version}", "Setup Mode"])

if __name__ == "__main__":
    main()